//---------------------------------------------------------------------------------------
// Датчик RMS на Arduino Pro Mini 5V/16MHz
// Файл rms_03 v.16.10.16.09.42
// OldBean
//---------------------------------------------------------------------------------------
// Краткое описание процесса синхронизации задач.
//---------------------------------------------------------------------------------------
// Параллельно выполняются следующие задачи:
// 1. Периодическая оцифровка входного сигнала (выпрямленное диодным мостиком напряжение сети,
//    пониженное до необходимого уровня трансформатором напряжения и резистивным делителем.
// 2. По окончанию преобразования производится возведение результата в квадрат и суммирование
//    с предыдущими значениями. После накопления необходимого для усреднения количества
//    результатов, производится деление суммы на количество просуммированных результатов и
//    извлечение квадратного корня.
// 3. Параллельно с процессами 1 и 2 производится регенерация трезразрядного семичегментного
//    индикатора для визуализации среднеквадратичного напряжения сети.
//
// Для устранения коллизий все эти процессы жестко синхронизированы. Тактирование работы 
// производится таймером TIMER2 - каждые 200 мкс происходят прерывания по совпадению c OCR2A.
// В процедуре обработки этого прерывания производится запуск АЦП. Частота преобразования АЦП
// установлена на 125 кГц, что вполне достаточно для 10-битного преобразования. Время 
// преобразования колеблется в пределах от 100 до 120 мкс (округленно). 
// 
// После 15 прерываний таймера (т.е. приблизительно раз в 3 мс) производится вызов функции 
// ShowSegments(), осуществляющей переход к подстветке следующей группы сегментов. Время 
// работы этой функции составляет около 28 мкс. Таким образом, переключение подсветки сегментов 
// происходит параллельно с циклом преобразования АЦП и никогда не выходит за его пределы.
// 
// По окончанию цикла преобразования АЦП происходит прерывание от АЦП. До начала следующего цикла
// преобразования АЦП остается приблизительно 80 - 100 мкс. За это время, в процедуре обработки
// этого прерывания от таймера, производится возведение результата оцифровки в квадрат и 
// суммирование этих квадратов. Здесь же, после суммирования 2048 значений (это соответствует 
// 41 полупериоду колебаний сети 50 Гц) производится деление деление суммы на 2048 (сдвиг 
// вправо на 11 бит) и извлечение квадратного корня из результата деления. Корень извлекается
// за время порядка 42 мкс, что вполне укладывается во временной интервал, остающийся до начала
// следующего цикла преобразования АЦП.
//---------------------------------------------------------------------------------------
#define DEBUG

#include <Wire.h>
#define SLAVE_ADDRESS       0x05      // Адрес шины i2c для связи с внешним миром. Slave

static uint8_t adcChannel = 3; // Входной аналоговый сигнал подключен к пину A3
// Установки, связанные с индикатором, рассчитаны на трехразрядные семисегментные индикаторы
// с общими катодами, подключенные: непосредственно (выводы анодов) и через резисторы 470 Ом
// (выводы общих катодов) к соответствующим пинам МК.
// Пины, к которым подключены сегменты индикатора (аноды)
static uint8_t segPins[7] = {12, 16, 5, 7, 8, 13, 4};
// Пины, к которым подключены общие катоды цифр индикатора
static uint8_t digPins[3] = {15, 14, 11};
static uint8_t digit[11] = {0b00111111, // Маски сегментов цифр 0, 1, 2 и т.д.
                            0b00000110,
                            0b01011011,
                            0b01001111,
                            0b01100110,
                            0b01101101,
                            0b01111101,
                            0b00000111,
                            0b01111111,
                            0b01101111,
                            0b00000000}; // "Цифра" 10 это пробел
static int d[3]; // Сотни, десятки и единицы показываемые на разрядах 7-сегментного индикатора
static uint8_t mask = 0x01; // Маска для анализа цифр
static int seg = 0; // Номер сегмента, обрабатываемого (подствечиваемого, если нужно) в данный момент

static volatile long rms = 137; // Среднеквадратичное значение измеряемого напряжения

// Рабочие переменные для отладки
//static volatile unsigned long stg = 0;

static volatile int cntr = 0; // Счетчик для тактирования смены подствечиваемых сегментов

static volatile unsigned long sum = 0; // Сумматор квадратов напряжения
static volatile int scntr = 0; // Счетчик квадратов прибавленных к сумме
//---------------------------------------------------------------------------------------
void setup() {
  for(int i = 0; i < 7; i++) { // Инициализация пинов, к которым подключены аноды индикатора (сегменты)
    pinMode(segPins[i], OUTPUT);
    digitalWrite(segPins[i], LOW);
  }
  for(int i = 0; i < 3; i++) { // Инициализация пинов, к которым подключены катоды индикатора (цифры)
    pinMode(digPins[i], OUTPUT);
    digitalWrite(digPins[i], HIGH);
  }
  unsigned long st = millis() + 1000;
  d[0] = 8; d[1] = 8; d[2] = 8; // Для демонстрации работоспособности индикатора покажем все сегменты разом
  while(millis() < st) {
      ShowSegments();
  }
  d[0] = 10; d[1] = 10; d[2] = 10; // Очистим индикатор
  ShowSegments();

#ifdef DEBUG
  pinMode(9, OUTPUT); // 9 - это PB1. Здесь используется для измерения временных характеристик МК
  digitalWrite(9, LOW);
//  PORTB |= 0b00000010; // Фронт импульса на пине 9
// .... фрагмент кода, длительность выполнения которого измеряется .....
//  PORTB &= 0b11111101; // Спад импульса на пине 9
#endif

  // Инициализация АЦП
  ADMUX = (1 << REFS1) | (1 << REFS0) |  // 11 - опорное напряжение = 1.1В
          (adcChannel & 0b00001111); // Канал АЦП соответствует пину, к которому поключен входной сигнал
  ADCSRA = (1 << ADEN) | // АЦП включен
           (0 << ADATE) | // Режим однократного преобразования
           (1 << ADIE) | // Прерывания по окончанию преобразования разрешены
           (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); // Предделитель - 128

  // Инициализация таймера 2 для тактирования АЦП и посветки индикатора
  TCCR2A = (1<<WGM21);            // Режим CTC (сброс по совпадению)
  TCCR2B = (1<<CS21) | (1<<CS20); // Предделитель 32
  OCR2A = 98;                     // При этом значении совпадения происходят с частотой около 5 кГц.
  TIMSK2 = (1<<OCIE2A);           // Разрешить прерывание по совпадению
}
//---------------------------------------------------------------------------------------
ISR (TIMER2_COMPA_vect) { // Обработчик прерывания от таймера TIMER2
#ifdef DEBUG
  PORTB |= 0b00000010;
#endif
  ADCSRA |= (1 << ADSC); // Запуск АЦП
  if(++cntr >= 15) { // Устанавливаем флаг для перехода к посветке следующей группы сегментов
    cntr = 0; ShowSegments(); // Подсвечиваем очередную группу сегментов
  }
}
//---------------------------------------------------------------------------------------
ISR(ADC_vect) { // Обработчик перрывания от АЦП по окончаниии преобразования
#ifdef DEBUG
  PORTB &= 0b11111101;
#endif
  int adcVal = ADCL; adcVal += ADCH << 8;
  sum += (unsigned long)adcVal*adcVal;
  if(++scntr == 2048) { // Накопление квадратов напряжения закончено
//    PORTB |= 0b00000010;
    rms = sqrt(sum >> 11); // Делим на 2048 и извлекаем корень (около 42 мкс)
//    PORTB &= 0b11111101;
    sum = 0; scntr = 0;
    UpdateDigits(rms);
  }
}
//---------------------------------------------------------------------------------------
void loop() {  
}
//---------------------------------------------------------------------------------------
void UpdateDigits(int val) { // Подготовка данных (цифр) для визуализации
  d[2] = val/100;
  d[1] = (val - d[2]*100)/10;
  d[0] = val - d[2]*100 - d[1]*10;
  if(d[2] == 0) { // Не отображаем ведущие нули ежели что... ("Цифра" 10 - это "пустая" цифра, все ее сегменты потушены)
    d[2] = 10;
    if(d[1] == 0) d[1] = 10;
  }
}
//---------------------------------------------------------------------------------------
void ShowSegments() { // Подсвечивание очередной группы сегментов
  digitalWrite(segPins[seg], LOW); // "Роняем" потенциал анода "отработавшего" сегмента
  if(++seg < 7) mask <<= 1; // Переходим к следующему сегменту
  else { // Дошли до последнего сегмента
    seg = 0; mask = 0x01;
  }
  digitalWrite(segPins[seg], HIGH); // "Подбрасываем" потенциал анода нового текущего сегмента
  for(int bi = 0; bi < 3; bi++) { // Теперь пробегаем по цифрам (разрядам) индикатора
    if(digit[d[bi]] & mask) // Если данный сегмент в bi-ом разряде должен гореть,
      digitalWrite(digPins[bi], LOW); // то уроним потенциал катода соответствующей цифры (разряда)
    else // А если гореть не должен, то
      digitalWrite(digPins[bi], HIGH); // потенциал катода bi-й цифры должен быть высокий 
  }
}
//---------------------------------------------------------------------------------------
void receiveData(int byteCount) { // callback для принимаемых по i2c байтов
  while(Wire.available()) {
    Wire.read(); // Просто затычка - принимать-то нечего
  }
}
//---------------------------------------------------------------------------------------
void sendData() { // callback для передаваемых по i2c байтов
    Wire.write(rms); // Отдаем текущее значение среднеквадратичного напряжения
}
//---------------------------------------------------------------------------------------

